from typing import Tuple
from langchain_core.tools import tool
import subprocess
from ..utls import log, exec_sh_cmd


@tool
def run_bash_cmd(cmd: str) -> Tuple[str, str]:
    """Runs a bash command.

    Args:
      cmd: The cmd to exectute.

    Returns:
      Tuple[str, str]: A tuple, with the first value being the output and the second the error, if there is one.
    """

    return exec_sh_cmd(cmd)



from .chroma_utls import *
from .config_loader import *
from .logger_utls import log
from .model_factory import *
from .io import *
from .sh import *
from .env import *
from .retriever_factory import *
from .hash import *



from environs import Env
from typing import Optional
from os import environ


class EnvVarProvider:
    def __init__(self):
        self._env = Env()
        self._env.read_env(".env")

    def get_env_var(self, key: str, default: Optional[str] = None) -> str:

        if key in environ:
            return environ[key]

        val = self._env(key, default=None)
        if not val:
            return default

        return val



from typing import Awaitable
from ..core.utls import log, traverse_folder


async def embed_file_system_contents(file_system_path: str) -> Awaitable:
    """Embeds the contents of a file system.

    Args:
      file_system_path: The path to the file system to embed.

    Returns:
      Awaitable: an empty promise.
    """

    log(f"embed_file_system_contents START. file_system_path: {file_system_path}")

    file_tree = traverse_folder(
        folder_path=file_system_path,
        ignore_folders=[".git", "__pycache__", "node_modules", "bin", "obj"],
    )

    log(f"file_tree: {file_tree}")


async def embed_repo(repo_name: str, repo_target_dir: str) -> Awaitable:
    log(f"{embed_repo.__name__} START.")

    ignore_folders = ["node_modules", ".git", "bin", "obj", "__pycache__"]
    ignore_extensions = [
        ".pfx",
        ".crt",
        ".pem",
        ".postman_collection.json",
        ".postman_environment",
        ".png",
        ".gif",
        ".jpeg",
        ".jpg",
        ".ico",
        ".svg",
        ".woff",
        ".woff2",
        ".ttf",
        ".gz",
        ".zip",
        ".tar",
        ".tgz",
        ".tar.gz",
        ".rar",
        ".7z",
        ".pdf",
        ".doc",
        ".docx",
        ".xls",
        ".xlsx",
        ".ppt",
        ".pptx"
    ]

    file_dict = traverse_folder(repo_target_dir, ignore_folders, ignore_extensions)
    file_paths = [f"{k}/{f}" for k, v in file_dict.items() for f in v]
    log(f"{embed_repo.__name__} -> file_paths: {file_paths}")

    chroma_client = ChromaHttpClientFactory.create_with_auth()
    embedding_function = EmbeddingFactory.create()

    chunk_embed_and_publish(
        file_paths=file_paths,
        collection_name=repo_name,
        embedding_function=embedding_function,
        chroma_client=chroma_client,
        chunk_size=4000,
        chunk_overlap=200,
    )

    log(f"{embed_repo.__name__} END.")



from typing import List, Any
from langchain.agents import AgentExecutor, create_openai_tools_agent
from langchain_core.messages import BaseMessage, HumanMessage
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_openai import AzureChatOpenAI


def create_assistant_agent(llm: AzureChatOpenAI, tools: list, system_message: str):
    """Create an agent."""
    prompt = ChatPromptTemplate.from_messages(
        [
            (
                "system",
                "You are a helpful AI assistant, collaborating with other assistants."
                " Use the provided tools to progress towards answering the question."
                " If you or any of the other assistants have the final answer or deliverable,"
                " prefix your response with FINAL ANSWER so the team knows to stop."
                " You have access to the following tools: {tool_names}.\n{system_message}",
            ),
            MessagesPlaceholder(variable_name="messages"),
        ]
    )
    prompt = prompt.partial(system_message=system_message)
    prompt = prompt.partial(tool_names=", ".join([tool.name for tool in tools]))
    return prompt | llm.bind_tools(tools)


def create_agent(llm: AzureChatOpenAI, system_message: str, tools: List[Any] = []):
    """Create an agent."""
    prompt = ChatPromptTemplate.from_messages(
        [
            (
                "system",
                (
                    "{system_message}"
                    + "\n You have access to the following tools: {tool_names}."
                    if len(tools)
                    else ""
                ),
            ),
            MessagesPlaceholder(variable_name="messages"),
        ]
    )
    prompt = prompt.partial(system_message=system_message)

    if len(tools):
        prompt = prompt.partial(tool_names=", ".join([tool.name for tool in tools]))
        # return prompt | llm.bind_tools(tools)

    return prompt | llm



import yaml
import json
from typing import Annotated, Type, Any, List, Dict
from pydantic import BaseModel, Field

from langchain_chroma import Chroma
from langchain_text_splitters import RecursiveCharacterTextSplitter
from langchain_openai import AzureOpenAIEmbeddings
from langchain_community.document_loaders import TextLoader
from langchain.tools.retriever import create_retriever_tool
from langchain_core.tools import BaseTool
from langchain_core.vectorstores import VectorStoreRetriever

from ..utls import RetrieverFactory, log


class RetrieveAdditionalContextToolSchema(BaseModel):
    query: str = Field(description="The query to search for.")


class RetrieveAdditionalContextTool(BaseTool):
    """
    This tool queries a vectorstore for additional information.
    """

    name: str = "retrieve_additional_context"
    description: str = "Queries a vectorstore for additional information."
    args_schema: Type[BaseModel] = RetrieveAdditionalContextToolSchema
    retriever: Any = None

    def __init__(
        self,
        retriever: VectorStoreRetriever = None,
        /,
        **data: Any,
    ):
        """
        An agent tool for fetching addition context from a vectorstore.
        """

        super().__init__(**data)

        self.retriever = retriever
        if self.retriever is None:
            self.retriever = RetrieverFactory.create(collection_name=kwargs["collection_name"])

    def _run(self, query: str) -> str:
        """Use the tool"""
        log(f"{self.name}. START: query: {query}")
        resp = self.retriever.invoke(query)
        log(f"{self.name}. END: query: {query}")
        return resp

    async def _arun(self, query: str) -> str:
        """Use the tool"""
        return self._run(query)



from typing import Dict, Any, List
import chromadb
import uuid
from chromadb.config import Settings
from langchain_chroma import Chroma
from langchain_openai import AzureOpenAIEmbeddings
from langchain_community.document_loaders import TextLoader
from ..utls import (
    chunk_embed_and_publish,
    create_retriever,
    ChromaHttpClientFactory,
    EmbeddingFactory,
)


def init_chroma_client_data(file_data_hash: Dict[str, List[str]]):

    chroma_client = ChromaHttpClientFactory.create()
    azure_embedding = EmbeddingFactory.create()

    for collection_name, file_paths in file_data_hash.items():

        chunk_embed_and_publish(
            file_paths=file_paths,
            collection_name=collection_name,
            embedding_function=azure_embedding,
            chroma_client=chroma_client,
        )



import logging
from rich.logging import RichHandler

formatter = logging.Formatter(
    "[bold cyan]%(asctime)s[/] - [bold %(levelname)s]%(levelname)s[/] - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

rich_handler = RichHandler(markup=True)
rich_handler.setFormatter(formatter)

logger = logging.getLogger("colorful_logger")
logger.setLevel(logging.INFO)
logger.addHandler(rich_handler)


def log(msg):
    logger.info(msg)



import os
from typing import List, Dict, Any, Optional
from langchain_community.document_loaders import TextLoader
from langchain_text_splitters import RecursiveCharacterTextSplitter

from .env import EnvVarProvider
from .logger_utls import log


DEFAULT_CHUNK_SIZE = 1500
DEFAULT_CHUNK_OVERLAP = 50

env = EnvVarProvider()


def traverse_folder(
    folder_path: str, ignore_folders: List[str], ignore_extensions: List[str] = None
) -> Dict[str, List[str]]:

    log(f"{traverse_folder.__name__} START. folder_path: {folder_path}")

    file_dict = {}

    for root, dirs, files in os.walk(folder_path):
        dirs[:] = [d for d in dirs if d not in ignore_folders]

        if ignore_extensions:
            files = [
                f
                for f in files
                if not any(f.endswith(ext) for ext in ignore_extensions)
            ]

        file_dict[root] = files

    log(f"{traverse_folder.__name__} END.")

    return file_dict


def chunk_files(
    file_paths: List[str],
    chunk_size: Optional[int] = None,
    chunk_overlap: Optional[int] = None,
) -> Dict[str, Dict[str, Any]]:
    chunk_size = chunk_size or env.get_env_var("CHUNK_SIZE", DEFAULT_CHUNK_SIZE)
    chunk_overlap = chunk_overlap or env.get_env_var("CHUNK_OVERLAP", DEFAULT_CHUNK_OVERLAP)

    chunk_hash = {}

    for file_path in file_paths:
        loader = TextLoader(file_path)
        docs = loader.load()

        text_splitter = RecursiveCharacterTextSplitter.from_tiktoken_encoder(
            chunk_size=chunk_size, chunk_overlap=chunk_overlap
        )

        split_docs = text_splitter.split_documents(docs)
        split_texts = [doc.page_content for doc in split_docs]

        if not len(split_texts):
            continue

        chunk_hash[file_path] = {"split_docs": split_docs, "split_texts": split_texts}

    return chunk_hash



import yaml
import json
from typing import Annotated, Type, Any, List, Dict
from pydantic import BaseModel, Field

from langchain_chroma import Chroma
from langchain_text_splitters import RecursiveCharacterTextSplitter
from langchain_openai import AzureOpenAIEmbeddings
from langchain_community.document_loaders import TextLoader
from langchain.tools.retriever import create_retriever_tool
from langchain_core.tools import BaseTool
from langchain_core.vectorstores import VectorStoreRetriever

from ..utls import create_retriever, ChromaHttpClientFactory, log, EmbeddingFactory


class RetrieverFactory:

    @staticmethod
    def create(collection_name: str, /, **kwargs: Dict[str, Any]) -> VectorStoreRetriever:
        """
        A vector store retriever factory.
        """

        chroma_client = kwargs.get("chroma_client", None)
        if not chroma_client:
            chroma_client = ChromaHttpClientFactory.create()

        embedding_function = kwargs.get("embedding_function", None)
        if not embedding_function:
            embedding_function = EmbeddingFactory.create()

        retriever = create_retriever(
            collection_name=collection_name,
            chroma_client=chroma_client,
            embedding_function=embedding_function,
        )

        return retriever



import pprint
import functools
import operator
from typing import Type, Optional, List

from langchain_core.tools import Tool
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langgraph.graph import START, END, StateGraph
from langgraph.prebuilt.tool_executor import ToolExecutor

from ..utls import log, ModelFactory 
from ..agnts import (
    create_agent_executor,
    should_invoke_tools,
    invoke_tools,
    AgentState,
)


def build_agnt_with_tools_graph(
        sys_prompt: str, 
        tools: List[Tool], 
        state_type: Optional[Type] = AgentState, 
        agnt_name: Optional[str] = "agent"
):

    tool_executor = ToolExecutor(tools)

    prompt = ChatPromptTemplate.from_messages(
        [("system", sys_prompt), MessagesPlaceholder(variable_name="messages")]
    )

    model = ModelFactory.create().bind_tools(tools)
    chain = prompt | model
    agent_node = create_agent_executor(chain=chain)

    graph = StateGraph(state_type)

    graph.add_node(agnt_name, agent_node)
    graph.add_node(
        "invoke_tools", functools.partial(invoke_tools, tool_executor=tool_executor)
    )

    graph.add_edge(START, agnt_name)

    graph.add_conditional_edges(
        agnt_name,
        should_invoke_tools,
        {
            "invoke_tools": "invoke_tools",
            "continue": END,
        },
    )

    graph.add_edge("invoke_tools", agnt_name)
    graph.add_edge(agnt_name, END)

    return graph.compile()



from typing import Optional
from json import load as json_load
import os
from .env import EnvVarProvider


DEFAULT_CONFIG_FILE_PATH = ".config/openai_config.json"
env = EnvVarProvider()


def load_openai_config(config_file_path: Optional[str] = None):
    if config_file_path is None:
        config_file_path = env.get_env_var("CONFIG_PATH", DEFAULT_CONFIG_FILE_PATH)

    with open(config_file_path) as f:
        openai_config = json_load(f)
        os.environ["AZURE_OPENAI_API_KEY"] = openai_config["openai_api_key"]
        os.environ["AZURE_OPENAI_ENDPOINT"] = openai_config["azure_endpoint"]
        os.environ["AZURE_OPENAI_API_VERSION"] = openai_config["openai_api_version"]
        return openai_config



from typing import List, Optional
import chromadb
import uuid
import base64
from chromadb.config import Settings
from langchain_chroma import Chroma
from langchain_openai import AzureOpenAIEmbeddings
from langchain_text_splitters import RecursiveCharacterTextSplitter
from .model_factory import EmbeddingFactory
from .env import EnvVarProvider
from .io import chunk_files


DEFAULT_HOST = "localhost"
DEFAULT_PORT = 8000
DEFAULT_USR = "admin"
DEFAULT_PWD = "admin"
DEFAULT_CHUNK_SIZE = 1500
DEFAULT_CHUNK_OVERLAP = 50

env = EnvVarProvider()


class ChromaHttpClientFactory:
    @staticmethod
    def create_with_auth_header():
        host = env.get_env_var("CHROMA_HOST", DEFAULT_HOST)
        port = env.get_env_var("CHROMA_PORT", DEFAULT_PORT)
        usr = env.get_env_var("CHROMA_USR", DEFAULT_USR)
        pwd = env.get_env_var("CHROMA_PWD", DEFAULT_PWD)

        auth_str = f"{usr}:{pwd}"
        encoded_auth = base64.b64encode(auth_str.encode()).decode()
        headers = {"Authorization": f"Basic {encoded_auth}"}

        chroma_client = chromadb.HttpClient(
            settings=Settings(allow_reset=True), host=host, port=port, headers=headers
        )

        return chroma_client


    @staticmethod
    def create_with_auth():
        host = env.get_env_var("CHROMA_HOST", DEFAULT_HOST)
        port = env.get_env_var("CHROMA_PORT", DEFAULT_PORT)
        usr = env.get_env_var("CHROMA_USR", DEFAULT_USR)
        pwd = env.get_env_var("CHROMA_PWD", DEFAULT_PWD)

        auth_str = f"{usr}:{pwd}"

        chroma_client = chromadb.HttpClient(
            settings=Settings(allow_reset=True, chroma_client_auth_provider="chromadb.auth.basic_authn.BasicAuthClientProvider", chroma_client_auth_credentials=auth_str), 
            host=host, 
            port=port
        )

        return chroma_client


    @staticmethod
    def create():
        host = env.get_env_var("CHROMA_HOST", DEFAULT_HOST)
        port = env.get_env_var("CHROMA_PORT", DEFAULT_PORT)

        chroma_client = chromadb.HttpClient(
            settings=Settings(allow_reset=True), host=host, port=port
        )

        return chroma_client


def embed_and_publish(file_paths, collection_name):
    documents = []
    collection = chroma_client.create_collection(collection_name)

    for file_path in file_paths:
        with open(file_path, "r", encoding="utf-8") as file:
            content = file.read()
            embedding = azure_embedding.embed_documents([content])[0]
            documents.append(
                {
                    "id": file_path,
                    "text": content,
                    "embedding": embedding,
                }
            )

    collection.add(documents=documents)


def chunk_embed_and_publish(
    file_paths: List[str],
    collection_name: str,
    embedding_function: AzureOpenAIEmbeddings,
    chroma_client: chromadb.HttpClient,
    chunk_size: Optional[int] = None,
    chunk_overlap: Optional[int] = None,
):
    chunk_size = chunk_size or env.get_env_var("CHUNK_SIZE", DEFAULT_CHUNK_SIZE)
    chunk_overlap = chunk_overlap or env.get_env_var("CHUNK_OVERLAP", DEFAULT_CHUNK_OVERLAP)

    vector_store = Chroma(
        embedding_function=embedding_function,
        client=chroma_client,
        collection_name=collection_name,
    )

    chunked_file_hash = chunk_files(file_paths, chunk_size, chunk_overlap)

    for file_path in chunked_file_hash:

        split_docs = chunked_file_hash[file_path]["split_docs"]
        split_texts = chunked_file_hash[file_path]["split_texts"]

        embeddings = embedding_function.embed_documents(split_texts)
        ids = [f"{file_path}_{i}" for i in range(len(embeddings))]

        if not len(ids):
            continue

        vector_store.add_documents(documents=split_docs, embeddings=embeddings, ids=ids)


def create_retriever(
    collection_name: str, chroma_client: chromadb.HttpClient, embedding_function: AzureOpenAIEmbeddings
):
    vector_store = Chroma(
        embedding_function=embedding_function,
        collection_name=collection_name,
        client=chroma_client,
    )

    # retriever = vector_store.as_retriever(
    #     search_type="similarity", search_kwargs={"k": 5}
    # )
    retriever = vector_store.as_retriever()

    return retriever
